#!/usr/bin/python3

######
## This script takes a json file of paired issues and translates the attachment and comment ids to the migrated system.
######

import requests, json, sys, getopt, re
from datetime import datetime, timezone

DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"

### parsing command line arguments
def getOptions(argv):
	def printUsage():
		print('test.py -f <mappingFile> --bu1 <urlToFirstJira> --u1 <userFirstJira> --p1 <passwordFirstJira> --bu2 <urlToSecondJira> --u2 <userSecondJira> --p2 <passwordSecondJira>')
		print('If the lookup needs to be performed on the secondConnector, please specify the "-r" option for reversing the connectors.')
		print('If you need to use bearer token, please specify the "-b1" or "-b2" option.')

	try:
		opts, args = getopt.getopt(argv,"hrf:b1b2",["file=","bu1=","u1=","p1=","bu2=","u2=","p2=", "baseUrl1=","user1=","password1=","baseUrl2=","user2=","password2=","b1","b2"])
	except getopt.GetoptError as e:
		print(str(e))
		printUsage()
		sys.exit(2)

	reverseConnectors = False
	useBearer1 = False
	useBearer2 = False

	for opt, arg in opts:
		if opt == '-h':
			printUsage()
			sys.exit()
		elif opt == '-r':
			reverseConnectors = True
		elif opt == '--b1':
			useBearer1 = True
		elif opt == '--b2':
			useBearer2 = True
		elif opt in ("-f", "--file"):
			mappingFile = arg
		elif opt in ("--bu1", "--baseUrl1"):
			baseUrl1 = arg
		elif opt in ("--u1", "--user1"):
			user1 = arg
		elif opt in ("--p1", "--password1"):
			password1 = arg
		elif opt in ("--bu2", "--baseUrl2"):
			baseUrl2 = arg
		elif opt in ("--u2", "--user2"):
			user2 = arg
		elif opt in ("--p2", "--password2"):
			password2 = arg
	if 'mappingFile' not in locals():
		print('mapping file is missing')
		printUsage()
		sys.exit(2)
	if 'baseUrl1' not in locals() or 'user1' not in locals() or 'password1' not in locals():
		print('connection information for first Jira is missing')
		printUsage()
		sys.exit(2)
	if 'baseUrl2' not in locals() or 'user2' not in locals() or 'password2' not in locals():
		print('connection information for second Jira is missing')
		printUsage()
		sys.exit(2)

	return mappingFile, reverseConnectors, baseUrl1, user1, password1, baseUrl2, user2, password2, useBearer1, useBearer2


### checking if the data for this connector contains deleted/not visible issues
def processConnector(connectorMappingData, baseUrl, user, password, baseUrl2, user2, password2, useBearer1, useBearer2):
	mappingsPerIssue = {}
	# iterate over firstConnector
	for issueData in connectorMappingData:
		try:
			localIssueKey = issueData.get('localIssueKey')
			print(f'processing {localIssueKey}')
			if localIssueKey:
				firstStatusCode, firstAttachmentsMap, firstCommentsMap = getAttachmentsAndComments(localIssueKey, getIdKey, getIdKey, baseUrl, user, password, useBearer1)
				if firstStatusCode == 200:
					secondStatusCode, secondAttachmentsMap, secondCommentsMap = getAttachmentsAndComments(localIssueKey, getFilenameCreatedKey, getBodyCreatedKey, baseUrl2, user2, password2, useBearer2)
					if secondStatusCode == 200:
						attachmentIdMapping = getAttachmentIdMapping(firstAttachmentsMap, secondAttachmentsMap, issueData)
						commentIdMapping = getCommentIdMapping(firstCommentsMap, secondCommentsMap, issueData)
						mappingsPerIssue[localIssueKey] = {'attachmentIdMapping': attachmentIdMapping, 'commentIdMapping': commentIdMapping}
		except Exception as e:
			print(f'Error processing {localIssueKey}: {e}')
	return mappingsPerIssue


def getAttachmentsAndComments(localIssueKey, attachmentKeyFunction, commentKeyFunction, baseUrl, user, password, useBearer):
	issueResponse = make_request(f"{baseUrl}/rest/api/2/issue/{localIssueKey}?fields=attachment,comment", user, password, useBearer)
	# print(f'response {issueResponse.text}')
	status_code = issueResponse.status_code

	if status_code == 200:
		issue = issueResponse.json()
		fields = issue.get('fields')
		if fields:
			attachmentsMap = {}
			for attachment in fields.get('attachment'):
				#print(f'attachment with id {attachment['id']} and name {attachment['filename']}')
				attachmentsMap[attachmentKeyFunction(attachment)] = attachment
			commentsMap = {}
			comment = fields.get('comment')
			if comment:
				for comment in comment.get('comments'):
					#print(f'comment with id {comment['id']} and name {comment['body']}')
					comment['body'] = comment['body'].replace('\r\n', '\n')
					comment['body'] = normalize_user_mentions(comment['body'])
					commentsMap[commentKeyFunction(comment)] = comment
		return status_code, attachmentsMap, commentsMap
	else:
		print(f'issue {localIssueKey} not found in {baseUrl}')
		return status_code, None, None

def convert_to_utc(timestamp):
	if not timestamp:
		return datetime.now(timezone.utc).strftime(DATE_FORMAT)
	try:
		dt = datetime.strptime(timestamp, DATE_FORMAT)
		dt_utc = dt.astimezone(timezone.utc)
		return dt_utc.strftime(DATE_FORMAT)
	except ValueError:
		return datetime.now(timezone.utc).strftime(DATE_FORMAT)

def getIdKey(attachmentOrComent):
	return attachmentOrComent['id']

def getFilenameCreatedKey(attachment):
	return f'{attachment["filename"]}-{convert_to_utc(attachment["created"])}'

def getBodyCreatedKey(comment):
	return f'{comment["body"]}-{convert_to_utc(comment["created"])}'

def getAttachmentIdMapping(firstAttachmentsMap, secondAttachmentsMap, issueData):
	attachmentIdMapping = {}
	# Iterate over attachments in firstAttachmentsMap
	for firstAttachmentId, firstAttachment in firstAttachmentsMap.items():
		filename = getFilenameCreatedKey(firstAttachment)
		# Check if filename exists in secondAttachmentsMap
		if filename in secondAttachmentsMap:
			secondAttachment = secondAttachmentsMap[filename]
			secondAttachmentId = secondAttachment['id']
			attachmentIdMapping[firstAttachmentId] = secondAttachmentId
		else:
			print(f'No mapping found for attachment with filename: {filename}')
			if 'attachments' in issueData:
				for attachment in issueData.get('attachments'):
					if 'localId' in attachment:
						localId = attachment.get('localId')
						if localId in attachmentIdMapping:
							print(f'Mapping attachment: {localId} --> {attachmentIdMapping[localId]}')
							attachment['localId'] = attachmentIdMapping[localId]
						else:
							print(f'No attachment found with localId: {localId}')
	return attachmentIdMapping

def getCommentIdMapping(firstCommentsMap, secondCommentsMap, issueData):
	commentIdMapping = {}
	# Iterate over comments in firstCommentsMap
	for firstCommentId, firstComment in firstCommentsMap.items():
		body = getBodyCreatedKey(firstComment)
		# Check if body exists in secondCommentsMap
		if body in secondCommentsMap:
			secondComment = secondCommentsMap[body]
			secondCommentId = secondComment['id']
			commentIdMapping[firstCommentId] = secondCommentId
		else:
			print(f'No mapping found for comment with body: {body}')
			if 'comments' in issueData:
				for comment in issueData.get('comments'):
					if 'localId' in comment:
						localId = comment.get('localId')
						if localId in commentIdMapping:
							print(f'Mapping comment: {localId} --> {commentIdMapping[localId]}')
							comment['localId'] = commentIdMapping[localId]
						else:
							print(f'No comment found with localId: {localId}')
	return commentIdMapping

def replaceIds(keyToReplace, issueKeyType, connectorMappingData, mappingsPerIssue):
	for issueData in connectorMappingData:
		issueKey = issueData.get(issueKeyType)
		print(f'processing {issueKey}')

		if issueKey in mappingsPerIssue:
			mappingPerIssue = mappingsPerIssue[issueKey]
			attachmentIdMapping = mappingPerIssue.get('attachmentIdMapping')
			commentIdMapping = mappingPerIssue.get('commentIdMapping')

			if 'attachments' in issueData:
				for attachment in issueData.get('attachments'):
					if keyToReplace in attachment:
						remoteId = attachment.get(keyToReplace)
						if remoteId in attachmentIdMapping:
							print(f'Mapping attachment: {remoteId} --> {attachmentIdMapping[remoteId]}')
							attachment[keyToReplace] = attachmentIdMapping[remoteId]
						else:
							print(f'No attachment found with id: {remoteId}')

			if 'comments' in issueData:
				for comment in issueData.get('comments'):
					if keyToReplace in comment:
						remoteId = comment.get(keyToReplace)
						if remoteId in commentIdMapping:
							print(f'Mapping comment: {remoteId} --> {commentIdMapping[remoteId]}')
							comment[keyToReplace] = commentIdMapping[remoteId]
						else:
							print(f'No comment found with id: {remoteId}')

def saveJson(jsonData, filename):
	print(f'saving json to {filename}')

	with open(filename, 'w') as f:
		json.dump(jsonData, f, indent=4)

def normalize_user_mentions(body):
	if not body:
		return body

	mention_pattern = r'\[~[^\]]+\]'
	normalized_body = re.sub(mention_pattern, '[~user]', body)

	return normalized_body

def make_request(url, user, password, useBearer):
	headers = {}
	if useBearer:
		headers['Authorization'] = f'Bearer {password}'
	else:
		headers['Authorization'] = requests.auth._basic_auth_str(user, password)
	response = requests.get(url, headers=headers)
	return response

### start of script
mappingFile, reverseConnectors, baseUrl1, user1, password1, baseUrl2, user2, password2, useBearer1, useBearer2 = getOptions(sys.argv[1:])

#parse mapping file
with open(mappingFile) as f:
	mappingData = json.load(f)
	#print(mappingData)

if reverseConnectors:
	connector1 = "secondConnector"
	connector2 = "firstConnector"
else:
	connector1 = "firstConnector"
	connector2 = "secondConnector"

print(f'### Processing sync infos for {connector1}')
mappingsPerIssue = processConnector(mappingData[connector1], baseUrl1, user1, password1, baseUrl2, user2, password2, useBearer1, useBearer2)

if connector2 in mappingData:
	print(f'### Replacing remote ids for {connector2}')
	replaceIds('remoteId', 'remoteIssueKey', mappingData[connector2], mappingsPerIssue)
	replaceIds('localId', 'localIssueKey', mappingData[connector1], mappingsPerIssue)

saveJson(mappingData, mappingFile + '-converted2.json')
